import atexit
import logging
import signal
from abc import ABC, abstractmethod
from datetime import datetime
from uuid import UUID

from hyrex import constants
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    DequeuedTask,
    EnqueueTaskRequest,
    TaskStatus,
    WorkflowRunRequest,
)


class Dispatcher(ABC):
    logger = logging.getLogger(__name__)

    def _signal_handler(self, signum, frame):
        signame = signal.Signals(signum).name
        self.logger.debug(f"\nReceived {signame}. Shutting down Hyrex dispatcher...")
        self.stop()

    def _chain_signal_handlers(self, new_handler, old_handler):
        """Return a function that calls both the new and old signal handlers."""

        def wrapper(signum, frame):
            # Call the new handler first
            new_handler(signum, frame)
            # Then call the previous handler (if it exists)
            if old_handler and callable(old_handler):
                old_handler(signum, frame)

        return wrapper

    def _setup_signal_handlers(self):
        for sig in (signal.SIGTERM, signal.SIGINT):
            old_handler = signal.getsignal(sig)  # Get the existing handler
            new_handler = self._signal_handler  # Your new handler
            # Set the new handler, which calls both new and old handlers
            signal.signal(sig, self._chain_signal_handlers(new_handler, old_handler))

    def register_shutdown_handlers(self):
        self._setup_signal_handlers()
        atexit.register(self.stop)

    def __init__(self):
        pass

    @abstractmethod
    def register_app(self, app_info: dict):
        pass

    @abstractmethod
    def enqueue(
        self,
        tasks: list[EnqueueTaskRequest],
    ):
        pass

    @abstractmethod
    def dequeue(
        self,
        executor_id: UUID,
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        pass

    @abstractmethod
    def mark_success(self, task_id: UUID):
        pass

    @abstractmethod
    def mark_failed(self, task_id: UUID):
        pass

    @abstractmethod
    def attempt_retry(self, task_id: UUID):
        pass

    @abstractmethod
    def try_to_cancel_task(self, task_id: UUID):
        pass

    @abstractmethod
    def task_canceled(self, task_id: UUID):
        pass

    # Result must be a JSON string
    @abstractmethod
    def save_result(self, task_id: UUID, result: str):
        pass

    @abstractmethod
    def get_task_status(self, task_id: UUID) -> TaskStatus:
        pass

    @abstractmethod
    def register_task(self, task_name: str, cron: str = None, source_code: str = None):
        pass

    @abstractmethod
    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        pass

    @abstractmethod
    def disconnect_executor(self, executor_id: UUID):
        pass

    @abstractmethod
    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        pass

    @abstractmethod
    def update_executor_stats(self, executor_id: UUID, stats: dict):
        pass

    @abstractmethod
    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        pass

    @abstractmethod
    def get_tasks_up_for_cancel(self) -> list[UUID]:
        pass

    @abstractmethod
    def get_queues_for_pattern(self, pattern: str) -> list[str]:
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def register_workflow(self, name: str, source_code: str, workflow_dag_json: str):
        pass

    @abstractmethod
    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        pass

    @abstractmethod
    def advance_workflow_run(self, workflow_run_id: UUID):
        pass



import threading
import time
from datetime import datetime
from queue import Empty, Queue
from uuid import UUID

import requests

from hyrex import constants
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    DequeuedTask,
    EnqueueTaskRequest,
    TaskStatus,
    WorkflowRunRequest,
)


class PlatformDispatcher(Dispatcher):

    HYREX_PLATFORM_URL = "https://platform-dev.hyrex.io"
    DEQUEUE_TASK_PATH = "/connect/dequeue-task"
    GET_STATUS_PATH = "/connect/get-task-status"
    UPDATE_STATUS_PATH = "/connect/update-task-status"
    ENQUEUE_TASK_PATH = "/connect/enqueue-task"

    def __init__(self, api_key: str, batch_size=100, flush_interval=0.1):
        super().__init__()
        self.api_key = api_key

        self.local_queue = Queue()
        self.running = True
        self.batch_size = batch_size
        self.flush_interval = flush_interval

        self.thread = threading.Thread(target=self._batch_enqueue, daemon=True)
        self.thread.start()

        self.register_shutdown_handlers()

    def register_app(self, app_info: dict):
        pass

    def enqueue(
        self,
        tasks: list[EnqueueTaskRequest],
    ):
        for task in tasks:
            self.local_queue.put(task)

    def _batch_enqueue(self):
        tasks = []
        last_flush_time = time.monotonic()
        while self.running or not self.local_queue.empty():
            timeout = self.flush_interval - (time.monotonic() - last_flush_time)
            if timeout <= 0:
                # Flush if the flush interval has passed
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()
                continue

            try:
                # Wait for a task or until the timeout expires
                task = self.local_queue.get(timeout=timeout)
                tasks.append(task)
                if len(tasks) >= self.batch_size:
                    # Flush if batch size is reached
                    self._enqueue_tasks(tasks)
                    tasks = []
                    last_flush_time = time.monotonic()
            except Empty:
                # No task received within the timeout
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()

        # Flush any remaining tasks when stopping
        if tasks:
            self._enqueue_tasks(tasks)

    # TODO: Add in all task fields once platform supports them
    def _enqueue_tasks(self, tasks: list[EnqueueTaskRequest]):
        enqueue_url = f"{self.HYREX_PLATFORM_URL}{self.ENQUEUE_TASK_PATH}"
        headers = {
            "x-project-api-key": self.api_key,
        }
        data = {
            "tasks": [
                {
                    "id": str(task.id),
                    # "root_id": str(task.root_id),
                    "task_name": task.task_name,
                    "queue": task.queue,
                    "args": task.args,
                    "max_retries": task.max_retries,
                    # "priority": task.priority,
                }
                for task in tasks
            ]
        }
        try:
            response = requests.post(enqueue_url, headers=headers, json=data)
            if response.status_code != 200:
                self.logger.error(f"Error enqueuing task: {response.status_code}")
                self.logger.error(f"Response body: {response.text}")
        except requests.exceptions.RequestException as e:
            self.logger.error(f"Error enqueuing task via API: {str(e)}")
            raise RuntimeError(f"Failed to enqueue task via API: {e}")

    def stop(self):
        """
        Stops the batching process and flushes remaining tasks.
        """
        self.logger.debug("Stopping dispatcher...")
        self.running = False
        self.thread.join()
        self.logger.debug("Dispatcher stopped successfully!")

    def dequeue(
        self,
        worker_id: UUID,
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        # TODO: Apply concurrency limit
        fetch_url = f"{self.HYREX_PLATFORM_URL}{self.DEQUEUE_TASK_PATH}"
        try:
            headers = {"x-project-api-key": self.api_key}
            dequeued_task = None
            json_body = {
                "queue": queue,
                "worker_id": str(worker_id),
                "num_tasks": 1,
            }
            response = requests.post(fetch_url, headers=headers, json=json_body)

            if response.status_code == 200:
                data = response.json()
                if data["tasks"]:
                    task = data["tasks"][0]
                    dequeued_task = DequeuedTask(
                        id=task["id"], name=task["task_name"], args=task["args"]
                    )

            else:
                self.logger.error(f"Error fetching task: {response.status_code}")
                error_body = response.text()  # Get the response body as text
                self.logger.error(f"Response body: {error_body}")

            return dequeued_task

        except Exception as e:
            self.logger.error(f"Exception while fetching task: {str(e)}")
            return None

    def _update_task_status(self, task_id: UUID, new_status: TaskStatus):
        update_task_url = f"{self.HYREX_PLATFORM_URL}{self.UPDATE_STATUS_PATH}"
        try:
            headers = {"x-project-api-key": self.api_key}
            data = {
                "task_updates": [
                    {"task_id": str(task_id), "updated_status": new_status}
                ]
            }
            response = requests.post(update_task_url, headers=headers, json=data)

            if response.status_code != 200:
                self.logger.error(f"Error updating task status: {response.status_code}")

        except Exception as e:
            self.logger.error(f"Exception while updating task status: {str(e)}")

    def mark_success(self, task_id: UUID):
        self._update_task_status(task_id, TaskStatus.success)

    def mark_failed(self, task_id: UUID):
        self._update_task_status(task_id, TaskStatus.failed)

    def attempt_retry(self, task_id: UUID):
        raise NotImplementedError("Retries not yet implemented on Hyrex platform")

    # TODO: Implement
    def try_to_cancel_task(self, task_id: UUID):
        raise NotImplementedError("Cancellation not yet implemented on Hyrex platform")

    def task_canceled(self, task_id: UUID):
        raise NotImplementedError("Cancellation not yet implemented on Hyrex platform")

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        get_status_url = f"{self.HYREX_PLATFORM_URL}{self.GET_STATUS_PATH}"
        try:
            headers = {"x-project-api-key": self.api_key}
            data = {"task_ids": [str(task_id)]}
            response = requests.get(get_status_url, headers=headers, json=data)

            if response.status_code == 200:
                response_data = response.json()
                if response_data.get("result"):
                    return response_data.get("result")[0].get("status")
                else:
                    raise ValueError(f"Task status not returned for task ID {task_id}")
            else:
                self.logger.error(f"Error getting task status: {response.status_code}")
                self.logger.error(f"Response body: {response.text}")
        except Exception as e:
            if isinstance(e, ValueError):
                raise
            self.logger.error(f"Exception while getting task status: {str(e)}")

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        pass

    def disconnect_executor(self, executor_id: UUID):
        pass

    def mark_running_tasks_lost(self, executor_id: UUID):
        pass

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        pass

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        pass

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        pass

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        pass

    def get_queues_for_pattern(self, pattern: str) -> list[str]:
        pass

    def register_task(self, task_name: str, cron: str = None, source_code: str = None):
        pass

    def register_workflow(self, name: str, source_code: str, workflow_dag_json: dict):
        pass

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        pass

    def advance_workflow_run(self, workflow_run_id: UUID):
        pass



import json
import random
import threading
import time
from contextlib import contextmanager
from datetime import datetime
from queue import Empty, Queue
from typing import List
from uuid import UUID

from psycopg import RawCursor
from psycopg.types.json import Json
from psycopg_pool import ConnectionPool
from uuid_extensions import uuid7

from hyrex import constants
from hyrex.dispatcher.dispatcher import Dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.schemas import (
    DequeuedTask,
    EnqueueTaskRequest,
    TaskStatus,
    WorkflowRunRequest,
    WorkflowStatus,
)
from hyrex.sql import sql, workflow_sql


class PostgresDispatcher(Dispatcher):
    def __init__(self, conn_string: str, batch_size=1000, flush_interval=0.05):
        super().__init__()
        self.conn_string = conn_string
        self.pool = ConnectionPool(
            conn_string + "?keepalives=1&keepalives_idle=60&keepalives_interval=10",
            open=True,
            max_idle=5,
        )

        self.local_queue = Queue()
        self.batch_size = batch_size
        self.flush_interval = flush_interval

        self.thread = threading.Thread(target=self._batch_enqueue, daemon=True)
        self.thread.start()
        self.stopping = False

        self.register_shutdown_handlers()

    @contextmanager
    def transaction(self):
        with self.pool.connection() as conn:
            with RawCursor(conn) as cur:
                try:
                    yield cur
                except InterruptedError:
                    conn.rollback()
                    raise
            conn.commit()

    def register_app(self, app_info: dict):
        with self.transaction() as cur:
            cur.execute(sql.REGISTER_APP_INFO_SQL, [1, app_info])

    def mark_success(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.MARK_TASK_SUCCESS, [task_id])

    def mark_failed(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.MARK_TASK_FAILED, [task_id])

    def attempt_retry(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(
                sql.CONDITIONALLY_RETRY_TASK,
                [task_id, uuid7()],
            )

    def try_to_cancel_task(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.TRY_TO_CANCEL_TASK, [task_id])

    def task_canceled(self, task_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.TASK_CANCELED, [task_id])

    def dequeue(
        self,
        executor_id: UUID,
        queue: str = constants.ANY_QUEUE,
        concurrency_limit: int = 0,
    ) -> DequeuedTask:
        dequeued_task = None
        with self.transaction() as cur:
            if concurrency_limit > 0:
                cur.execute(
                    sql.FETCH_TASK_WITH_CONCURRENCY,
                    [queue, concurrency_limit, executor_id],
                )
            else:
                cur.execute(sql.FETCH_TASK, [queue, executor_id])
            row = cur.fetchone()
            if row:
                (
                    task_id,
                    durable_id,
                    root_id,
                    parent_id,
                    task_name,
                    args,
                    queue,
                    priority,
                    timeout_seconds,
                    scheduled_start,
                    queued,
                    started,
                    workflow_run_id,
                ) = row
                dequeued_task = DequeuedTask(
                    id=task_id,
                    durable_id=durable_id,
                    root_id=root_id,
                    parent_id=parent_id,
                    task_name=task_name,
                    args=args,
                    queue=queue,
                    priority=priority,
                    timeout_seconds=timeout_seconds,
                    scheduled_start=scheduled_start,
                    queued=queued,
                    started=started,
                    workflow_run_id=workflow_run_id,
                )

        return dequeued_task

    def enqueue(self, tasks: list[EnqueueTaskRequest]):
        if not tasks:
            self.logger.error("Task list is empty - cannot enqueue.")
            return
        if self.stopping:
            self.logger.warning("Task enqueued during shutdown. May not be processed.")
        for task in tasks:
            self.local_queue.put(task)

    def _batch_enqueue(self):
        tasks = []
        last_flush_time = time.monotonic()
        while True:
            time_left = self.flush_interval - (time.monotonic() - last_flush_time)
            if time_left <= 0:
                # Flush if the flush interval has passed
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()
                continue

            try:
                # Wait for a task or until the timeout expires
                task = self.local_queue.get(timeout=time_left)
                if task is None:
                    # Stop sequence initiated
                    break
                tasks.append(task)
                if len(tasks) >= self.batch_size:
                    # Flush if batch size is reached
                    self._enqueue_tasks(tasks)
                    tasks = []
                    last_flush_time = time.monotonic()
            except Empty:
                # No task received within the timeout
                if tasks:
                    self._enqueue_tasks(tasks)
                    tasks = []
                last_flush_time = time.monotonic()

        # Flush any remaining tasks when stopping
        if tasks:
            self._enqueue_tasks(tasks)

    def _enqueue_tasks(self, tasks: List[EnqueueTaskRequest]):
        """
        Inserts a batch of tasks into the database.

        :param tasks: List of tasks to insert.
        """
        task_data = (
            (
                task.id,
                task.durable_id,
                task.root_id,
                task.parent_id,
                task.task_name,
                Json(task.args),
                task.queue,
                task.max_retries,
                task.priority,
                task.timeout_seconds,
                task.idempotency_key,
                task.status,
                task.workflow_run_id,
                task.workflow_dependencies,
            )
            for task in tasks
        )

        with self.transaction() as cur:
            cur.executemany(
                sql.ENQUEUE_TASK,
                task_data,
            )

    def stop(self, timeout: float = 5.0) -> bool:
        self.logger.debug("Stopping dispatcher...")
        self.stopping = True

        # Signal the batch thread to stop and wait with timeout
        self.local_queue.put(None)
        self.thread.join(timeout=timeout)

        clean_shutdown = not self.thread.is_alive()

        # Close the connection pool
        if clean_shutdown:
            self.pool.close()
        else:
            self.logger.warning(
                "Batch thread did not stop cleanly, forcing connection pool to close"
            )
            self.pool.close(timeout=1.0)

        self.logger.debug(
            "Dispatcher stopped %s.",
            "successfully" if clean_shutdown else "with timeout",
        )
        return clean_shutdown

    def get_task_status(self, task_id: UUID) -> TaskStatus:
        with self.transaction() as cur:
            cur.execute(sql.GET_TASK_STATUS, [task_id])
            result = cur.fetchone()
            if result is None:
                raise ValueError(f"Task id {task_id} not found in DB.")
            return result[0]

    def register_executor(
        self,
        executor_id: UUID,
        executor_name: str,
        queue_pattern: str,
        queues: list[HyrexQueue],
        worker_name: str,
    ):
        with self.transaction() as cur:
            cur.execute(
                sql.REGISTER_EXECUTOR,
                [executor_id, executor_name, queue_pattern, queues, worker_name],
            )

    def disconnect_executor(self, executor_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.DISCONNECT_EXECUTOR, [executor_id])

    def executor_heartbeat(self, executor_ids: list[UUID], timestamp: datetime):
        with self.transaction() as cur:
            cur.execute(sql.EXECUTOR_HEARTBEAT, [timestamp, executor_ids])

    def update_executor_stats(self, executor_id: UUID, stats: dict):
        with self.transaction() as cur:
            cur.execute(sql.UPDATE_EXECUTOR_STATS, [executor_id, stats])

    def task_heartbeat(self, task_ids: list[UUID], timestamp: datetime):
        with self.transaction() as cur:
            cur.execute(sql.TASK_HEARTBEAT, [timestamp, task_ids])

    def get_tasks_up_for_cancel(self) -> list[UUID]:
        with self.transaction() as cur:
            cur.execute(sql.GET_TASKS_UP_FOR_CANCEL)
            return [row[0] for row in cur.fetchall()]

    def mark_running_tasks_lost(self, executor_id: UUID):
        with self.transaction() as cur:
            cur.execute(sql.MARK_RUNNING_TASKS_LOST, [executor_id])

    def save_result(self, task_id: UUID, result: str):
        with self.transaction() as cur:
            cur.execute(sql.SAVE_RESULT, [task_id, result])

    def get_queues_for_pattern(self, pattern: str) -> list[str]:
        with self.transaction() as cur:
            cur.execute(sql.GET_QUEUES_FOR_PATTERN, [pattern])
            return [row[0] for row in cur.fetchall()]

    def register_task(self, task_name: str, cron: str = None, source_code: str = None):
        with self.transaction() as cur:
            cur.execute(sql.UPSERT_TASK, [task_name, cron, source_code])

    def register_workflow(self, name: str, source_code: str, workflow_dag_json: dict):
        with self.transaction() as cur:
            cron = None
            cur.execute(
                workflow_sql.UPSERT_WORKFLOW,
                [name, cron, source_code, json.dumps(workflow_dag_json)],
            )

    def send_workflow_run(self, workflow_run_request: WorkflowRunRequest) -> UUID:
        with self.transaction() as cur:
            cur.execute(
                workflow_sql.INSERT_WORKFLOW_RUN,
                [
                    workflow_run_request.id,
                    workflow_run_request.workflow_name,
                    Json(workflow_run_request.args),
                    workflow_run_request.queue,
                    workflow_run_request.timeout_seconds,
                    workflow_run_request.idempotency_key,
                ],
            )
            result = cur.fetchall()
            if len(result) != 1:
                raise ValueError(f"Insert workflow run failed.")
            return result[0][0]

    def advance_workflow_run(self, workflow_run_id: UUID):
        self.logger.info(f"Advancing workflow run {workflow_run_id}")

        with self.transaction() as cur:
            # First query to check status
            cur.execute(
                workflow_sql.SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS,
                [workflow_run_id],
            )
            result = cur.fetchall()

            if len(result) != 1:
                self.logger.warning(
                    "Result of SET_WORKFLOW_RUN_STATUS_BASED_ON_TASK_RUNS is not one row."
                )
                return None

            workflow_status = result[0][1]  # Status is second column
            if workflow_status in (WorkflowStatus.failed, WorkflowStatus.success):
                return None

            # Second query to advance the workflow
            cur.execute(workflow_sql.ADVANCE_WORKFLOW_RUN, [workflow_run_id])
            return None



import psycopg

from hyrex.sql import cron_sql, sql, stats_sql, workflow_sql


def create_tables(conn_string):
    with psycopg.connect(conn_string) as conn:
        with conn.cursor() as cur:
            cur.execute(sql.CREATE_HYREX_APP_TABLE)
            cur.execute(sql.CREATE_HYREX_TASK_RUN_TABLE)
            cur.execute(sql.CREATE_HYREX_TASK_TABLE)
            cur.execute(sql.CREATE_SYSTEM_LOG_TABLE)
            cur.execute(sql.CREATE_HYREX_RESULT_TABLE)
            cur.execute(sql.CREATE_HYREX_EXECUTOR_TABLE)
            cur.execute(cron_sql.CREATE_HYREX_CRON_JOB_TABLE)
            cur.execute(cron_sql.CREATE_HYREX_CRON_JOB_RUN_DETAILS_TABLE)
            cur.execute(cron_sql.CREATE_HYREX_SCHEDULER_LOCK_TABLE)
            cur.execute(cron_sql.CREATE_EXECUTE_QUEUED_COMMAND_FUNCTION)
            cur.execute(stats_sql.CREATE_HISTORICAL_TASK_STATUS_COUNTS)
            cur.execute(workflow_sql.CREATE_WORKFLOW_TABLE)
            cur.execute(workflow_sql.CREATE_WORKFLOW_RUN_TABLE)
        conn.commit()



from pydantic import BaseModel, Field

from hyrex.hyrex_queue import HyrexQueue


class TaskConfig(BaseModel):
    queue: str | HyrexQueue | None = None
    priority: int | None = Field(default=None, ge=0, le=10)
    max_retries: int | None = Field(default=None, ge=0)
    timeout_seconds: int | None = Field(default=None, gt=0)
    idempotency_key: str | None = None

    def get_queue_name(self) -> str:
        if isinstance(self.queue, str):
            return self.queue
        else:
            return self.queue.name

    def merge(self, other: "TaskConfig") -> "TaskConfig":
        """Merge another config into this one, with other taking precedence"""
        if other is None:
            return self

        merged_data = self.model_dump()
        other_data = other.model_dump()

        for field, value in other_data.items():
            if value is not None:
                merged_data[field] = value

        return TaskConfig(**merged_data)



import functools
import inspect
import logging
import os
from inspect import signature
from typing import Any, Callable

from hyrex import constants
from hyrex.config import EnvVars
from hyrex.dispatcher import Dispatcher, get_dispatcher
from hyrex.hyrex_queue import HyrexQueue
from hyrex.task import T, TaskWrapper
from hyrex.task_config import TaskConfig
from hyrex.workflow.workflow import HyrexWorkflow
from hyrex.workflow.workflow_builder import WorkflowBuilder

# TODO: Also register tasks with DB here.


class HyrexRegistry:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        if os.getenv(EnvVars.WORKER_PROCESS):
            self.dispatcher = None
        else:
            self.dispatcher = get_dispatcher()

        self.internal_task_registry: dict[str, TaskWrapper] = {}
        self.internal_queue_registry: dict[str, HyrexQueue] = {}

    def register_task(self, task_wrapper: TaskWrapper):
        self.logger.debug(f"Registering task: {task_wrapper.task_identifier}")
        if self.internal_task_registry.get(task_wrapper.task_identifier):
            raise KeyError(
                f"Task {task_wrapper.task_identifier} is already registered. Task names must be unique."
            )
        self.internal_task_registry[task_wrapper.task_identifier] = task_wrapper

        # Register the task wrapper's queue for tracking concurrency.
        queue = task_wrapper.get_queue()
        if isinstance(queue, str):
            self.register_queue(HyrexQueue(name=queue))
        else:
            self.register_queue(queue)

    def register_queue(self, queue: HyrexQueue):
        if self.internal_queue_registry.get(queue.name) and not queue.equals(
            self.internal_queue_registry[queue.name]
        ):
            raise KeyError(
                f"Conflicting concurrency limits on queue name: {queue.name}"
            )

        self.internal_queue_registry[queue.name] = queue

    def get_concurrency_limit(self, queue_name: str):
        # TODO: Add queue patterns?
        if self.internal_queue_registry.get(queue_name):
            return self.internal_queue_registry[queue_name].concurrency_limit
        else:
            return 0

    def set_dispatcher(self, dispatcher: Dispatcher):
        self.dispatcher = dispatcher
        for task_wrapper in self.internal_task_registry.values():
            task_wrapper.dispatcher = dispatcher

    def get_on_error_handler(self, task_name: str) -> Callable | None:
        task_wrapper = self.internal_task_registry[task_name]
        return task_wrapper.on_error

    def get_task_wrappers(self):
        return self.internal_task_registry.values()

    def get_task(self, task_name: str):
        return self.internal_task_registry[task_name]

    def add_registry(self, registry: "HyrexRegistry"):
        for task_wrapper in registry.get_task_wrappers():
            self.register_task(task_wrapper=task_wrapper)

    def task(
        self,
        func: Callable = None,
        *,
        queue: str | HyrexQueue = constants.DEFAULT_QUEUE,
        cron: str = None,
        max_retries: int = 0,
        timeout_seconds: int | None = None,
        priority: int = constants.DEFAULT_PRIORITY,
        on_error: Callable = None,
    ) -> TaskWrapper:
        """
        Create task decorator
        """

        def decorator(func: Callable[[T], Any]) -> TaskWrapper:
            task_identifier = func.__name__
            task_config = TaskConfig(
                queue=queue,
                max_retries=max_retries,
                timeout_seconds=timeout_seconds,
                priority=priority,
            )
            task_wrapper = TaskWrapper(
                task_identifier=task_identifier,
                func=func,
                cron=cron,
                task_config=task_config,
                dispatcher=self.dispatcher,
                on_error=on_error,
            )
            self.register_task(task_wrapper=task_wrapper)
            return task_wrapper

        if func is not None:
            return decorator(func)
        return decorator

    def schedule(self):
        for task_wrapper in self.values():
            task_wrapper.schedule()

    def workflow(
        self,
        name: str,
        queue: str | HyrexQueue = constants.DEFAULT_QUEUE,
        max_retries: int = 0,
        timeout_seconds: int | None = None,
        priority: int = constants.DEFAULT_PRIORITY,
        cron: str = None,
        workflow_arg_schema=None,
    ):
        """
        A decorator to register a workflow.
        """

        def decorator(func):
            task_config = TaskConfig(
                queue=queue,
                max_retries=max_retries,
                timeout_seconds=timeout_seconds,
                priority=priority,
            )

            with WorkflowBuilder() as workflow_builder:
                # Build the workflow by calling the function.
                func()

                # Register workflow on publisher (on worker processes, self.dispatcher won't be set yet)
                if self.dispatcher:
                    source_code = inspect.getsource(func)
                    self.dispatcher.register_workflow(
                        name=name,
                        source_code=source_code,
                        workflow_dag_json=workflow_builder.to_json(),
                    )

                # Create and return a HyrexWorkflow instance
                workflow = HyrexWorkflow(
                    name=name,
                    task_config=task_config,
                    workflow_arg_schema=workflow_arg_schema,
                    workflow_builder=workflow_builder,
                    dispatcher=self.dispatcher,
                )

            return workflow

        return decorator



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel


class HyrexContext(BaseModel):
    task_id: UUID
    durable_id: UUID
    root_id: UUID
    parent_id: UUID | None
    task_name: str
    queue: str
    priority: int
    timeout_seconds: int | None
    scheduled_start: datetime | None
    queued: datetime
    started: datetime
    executor_id: UUID


# Simple global context
_current_context: HyrexContext | None = None


def get_hyrex_context() -> HyrexContext | None:
    """Get the current Hyrex context."""
    return _current_context


def set_hyrex_context(context: HyrexContext) -> None:
    """Set the current Hyrex context."""
    global _current_context
    _current_context = context


def clear_hyrex_context() -> None:
    """Clear the current Hyrex context."""
    global _current_context
    _current_context = None
